/*
 * mod_gnustep_conf.c
 *
 * Apache module to configure Apache for using GNUstep
 *
 * Copyright (C) 2000 Free Software Foundation, Inc.
 *
 * Author:  Nicola Pero <n.pero@mi.flashnet.it>
 *
 * This file is part of GNUstep.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *  
 * You should have received a copy of the GNU General Public
 * License along with this library; see the file COPYING.LIB.
 * If not, write to the Free Software Foundation,
 * 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "httpd.h"
#include "http_config.h"
#include "http_log.h"
#include "ap_config.h"

/*
 * WHAT THIS MODULE DOES
 * =====================
 *
 * This module recognizes only the configuration directive
 *
 *  GNUstepRootPath root_path
 *
 * where root_path could for example be /usr/GNUstep, such as in:
 *
 * GNUstepRootPath /usr/local/GNUstep
 *
 * Upon encountering this directive, the module loads the GNUstep
 * environment into the server, as it is set by the
 * root_path/System/Makefiles/GNUstep.sh shell script.
 *
 *
 * HOW THIS MODULE WORKS
 * =====================
 *
 * It runs a child task.  The child task executes
 * root_path/System/Makefiles/PrintGNUstepEnv.sh, which is the
 * following shell script:

 #!/bin/sh
 source ${GNUSTEP_ROOT}/System/Makefiles/GNUstep.sh
 env

 * (of course we set GNUSTEP_ROOT before running the child task).
 * This script runs GNUstep.sh, and outputs the whole environment to
 * stdout.  We read the child stdout using a pipe, and change the
 * parent's environment (NB: the parent is the Apache server) according 
 * to the output of the child.  */


/*
 * Struct passed to the child task.
 */
struct gnustep_data {
  char *rootpath;
  struct pool *pool;
  server_rec *server;
};

/* This function is executed in a child task.  It executes a shell script 
 * which simply runs the GNUstep initialization script and then outputs 
 * to standard output all the environment using env.  The standard output 
 * is actually read by the parent using a pipe, and the parent sets its 
 * own environment to what is read from the pipe. */
static int output_gnustep_env (struct gnustep_data *data, 
			       child_info *info)
{
  char *gnustep_root = data->rootpath;
  struct pool *pool = data->pool;
  server_rec *server = data->server;
  char *script_path;

  if (gnustep_root == NULL)
    {
      gnustep_root = "/usr/GNUstep";
    }

  /* Set GNUSTEP_ROOT so that the PrintEnv.sh script can find the
     GNUstep GNUstep.sh init script */
  putenv (strdup (ap_pstrcat (pool, "GNUSTEP_ROOT=", gnustep_root, NULL)));
  
  /* FIXME: This should not be needed.  
     Makes sure path is enough to run grep, sed, env etc */
  /*  putenv (strdup (ap_pstrcat (pool, "PATH=", getenv ("PATH"), 
      ":/bin:/usr/bin", NULL)));*/
  
  /* Now build the shell script path */
  script_path = ap_pstrcat (pool, gnustep_root, 
			    "/System/Makefiles/PrintGNUstepEnv.sh", NULL);
  
  /* Execute the shell script */
  if (execl ("/bin/sh", "/bin/sh", "-c", script_path, NULL) < 0)
    {
      /* Could not execute it ! */
      ap_log_error (APLOG_MARK, APLOG_EMERG, server, 
		    "mod_gnustep_conf: Couldn't execute `%s'", script_path);
      exit (-1);
    }
  
  /* Never reached */
  return 0;
}

/*
 * Reads environment from stream, in the form used by the shell `env' 
 * command, breaks it into lines, and feed each line to putenv.
 */
static void read_gnustep_environment (FILE *stream)
{
  char *line = NULL;
  size_t n = 0;
  
  while (getline (&line, &n, stream) > -1)
    {
      putenv (strdup (line));
    }
  free (line);
}

/* Called when the GNUstepRootPath config directive is found */
static const char *set_rootpath (cmd_parms *cmd, void *dummy, char *rootpath)
{
  server_rec *server = cmd->server;
  struct pool *pool = cmd->pool;
  FILE *child_out;
  struct gnustep_data data_for_child;

  data_for_child.server = server;
  data_for_child.rootpath = rootpath;
  data_for_child.pool = pool;

  if (!ap_spawn_child 
      (pool, (int (*)(void *, child_info *)) output_gnustep_env, 
       (void *) &data_for_child, kill_after_timeout, NULL, &child_out, NULL))
    {
      ap_log_error (APLOG_MARK, APLOG_NOERRNO | APLOG_EMERG, server, 
		    "mod_gnustep_conf: couldn't spawn child process");
      return "Couldn't spawn child process";
    }
  
  read_gnustep_environment (child_out);

  return NULL;
}

/* Config file commands we recognize */
static const command_rec gnustep_conf_cmds[] =
{
  {"GNUstepRootPath", set_rootpath, NULL, RSRC_CONF, TAKE1,
   "the GNUSTEP_ROOT path (such as `/usr/GNUstep/')"},
  
  {NULL}
};

module MODULE_VAR_EXPORT gnustep_conf_module = {
    STANDARD_MODULE_STUFF, 
    NULL,                  /* module initializer                  */
    NULL,                  /* create per-dir    config structures */
    NULL,                  /* merge  per-dir    config structures */
    NULL,                  /* create per-server config structures */
    NULL,                  /* merge  per-server config structures */
    gnustep_conf_cmds,     /* table of config file commands       */
    NULL,                  /* [#8] MIME-typed-dispatched handlers */
    NULL,                  /* [#1] URI to filename translation    */
    NULL,                  /* [#4] validate user id from request  */
    NULL,                  /* [#5] check if the user is ok _here_ */
    NULL,                  /* [#3] check access by host address   */
    NULL,                  /* [#6] determine MIME type            */
    NULL,                  /* [#7] pre-run fixups                 */
    NULL,                  /* [#9] log a transaction              */
    NULL,                  /* [#2] header parser                  */
    NULL,                  /* child_init                          */
    NULL,                  /* child_exit                          */
    NULL                   /* [#0] post read-request              */
};

